# See https://en.wikipedia.org/wiki/General_MIDI#Parameter_interpretations
# Generated by get_instruments.py

instruments = {
    (1, "Acoustic Grand Piano", "piano"),
    (2, "Bright Acoustic Piano", "piano"),
    (3, "Electric Grand Piano", "piano"),
    (4, "Honky-tonk Piano", "piano"),
    (5, "Electric Piano 1", "piano"),
    (6, "Electric Piano 2", "piano"),
    (7, "Harpsichord", "piano"),
    (8, "Clavinet", "piano"),
    (9, "Celesta", "chromatic percussion"),
    (10, "Glockenspiel", "chromatic percussion"),
    (11, "Music Box", "chromatic percussion"),
    (12, "Vibraphone", "chromatic percussion"),
    (13, "Marimba", "chromatic percussion"),
    (14, "Xylophone", "chromatic percussion"),
    (15, "Tubular Bells", "chromatic percussion"),
    (16, "Dulcimer", "chromatic percussion"),
    (17, "Drawbar Organ", "organ"),
    (18, "Percussive Organ", "organ"),
    (19, "Rock Organ", "organ"),
    (20, "Church Organ", "organ"),
    (21, "Reed Organ", "organ"),
    (22, "Accordion", "organ"),
    (23, "Harmonica", "organ"),
    (24, "Tango Accordion", "organ"),
    (25, "Acoustic Guitar (nylon)", "guitar"),
    (26, "Acoustic Guitar (steel)", "guitar"),
    (27, "Electric Guitar (jazz)", "guitar"),
    (28, "Electric Guitar (clean)", "guitar"),
    (29, "Electric Guitar (muted)", "guitar"),
    (30, "Overdriven Guitar", "guitar"),
    (31, "Distortion Guitar", "guitar"),
    (32, "Guitar Harmonics", "guitar"),
    (33, "Acoustic Bass", "bass"),
    (34, "Electric Bass (finger)", "bass"),
    (35, "Electric Bass (pick)", "bass"),
    (36, "Fretless Bass", "bass"),
    (37, "Slap Bass 1", "bass"),
    (38, "Slap Bass 2", "bass"),
    (39, "Synth Bass 1", "bass"),
    (40, "Synth Bass 2", "bass"),
    (41, "Violin", "strings"),
    (42, "Viola", "strings"),
    (43, "Cello", "strings"),
    (44, "Contrabass", "strings"),
    (45, "Tremolo Strings", "strings"),
    (46, "Pizzicato Strings", "strings"),
    (47, "Orchestral Harp", "strings"),
    (48, "Timpani", "strings"),
    (49, "String Ensemble 1", "ensemble"),
    (50, "String Ensemble 2", "ensemble"),
    (51, "Synth Strings 1", "ensemble"),
    (52, "Synth Strings 2", "ensemble"),
    (53, "Choir Aahs", "ensemble"),
    (54, "Voice Oohs", "ensemble"),
    (55, "Synth Choir", "ensemble"),
    (56, "Orchestra Hit", "ensemble"),
    (57, "Trumpet", "brass"),
    (58, "Trombone", "brass"),
    (59, "Tuba", "brass"),
    (60, "Muted Trumpet", "brass"),
    (61, "French Horn", "brass"),
    (62, "Brass Section", "brass"),
    (63, "Synth Brass 1", "brass"),
    (64, "Synth Brass 2", "brass"),
    (65, "Soprano Sax", "reed"),
    (66, "Alto Sax", "reed"),
    (67, "Tenor Sax", "reed"),
    (68, "Baritone Sax", "reed"),
    (69, "Oboe", "reed"),
    (70, "English Horn", "reed"),
    (71, "Bassoon", "reed"),
    (72, "Clarinet", "reed"),
    (73, "Piccolo", "pipe"),
    (74, "Flute", "pipe"),
    (75, "Recorder", "pipe"),
    (76, "Pan Flute", "pipe"),
    (77, "Blown bottle", "pipe"),
    (78, "Shakuhachi", "pipe"),
    (79, "Whistle", "pipe"),
    (80, "Ocarina", "pipe"),
    (81, "Lead 1 (square)", "synth lead"),
    (82, "Lead 2 (sawtooth)", "synth lead"),
    (83, "Lead 3 (calliope)", "synth lead"),
    (84, "Lead 4 (chiff)", "synth lead"),
    (85, "Lead 5 (charang)", "synth lead"),
    (86, "Lead 6 (voice)", "synth lead"),
    (87, "Lead 7 (fifths)", "synth lead"),
    (88, "Lead 8 (bass + lead)", "synth lead"),
    (89, "Pad 1 (new age)", "synth pad"),
    (90, "Pad 2 (warm)", "synth pad"),
    (91, "Pad 3 (polysynth)", "synth pad"),
    (92, "Pad 4 (choir)", "synth pad"),
    (93, "Pad 5 (bowed)", "synth pad"),
    (94, "Pad 6 (metallic)", "synth pad"),
    (95, "Pad 7 (halo)", "synth pad"),
    (96, "Pad 8 (sweep)", "synth pad"),
    (97, "FX 1 (rain)", "synth effects"),
    (98, "FX 2 (soundtrack)", "synth effects"),
    (99, "FX 3 (crystal)", "synth effects"),
    (100, "FX 4 (atmosphere)", "synth effects"),
    (101, "FX 5 (brightness)", "synth effects"),
    (102, "FX 6 (goblins)", "synth effects"),
    (103, "FX 7 (echoes)", "synth effects"),
    (104, "FX 8 (sci-fi)", "synth effects"),
    (105, "Sitar", "ethnic"),
    (106, "Banjo", "ethnic"),
    (107, "Shamisen", "ethnic"),
    (108, "Koto", "ethnic"),
    (109, "Kalimba", "ethnic"),
    (110, "Bagpipe", "ethnic"),
    (111, "Fiddle", "ethnic"),
    (112, "Shanai", "ethnic"),
    (113, "Tinkle Bell", "percussive"),
    (114, "Agogo", "percussive"),
    (115, "Steel Drums", "percussive"),
    (116, "Woodblock", "percussive"),
    (117, "Taiko Drum", "percussive"),
    (118, "Melodic Tom", "percussive"),
    (119, "Synth Drum", "percussive"),
    (120, "Reverse Cymbal", "percussive"),
    (121, "Guitar Fret Noise", "sound effects"),
    (122, "Breath Noise", "sound effects"),
    (123, "Seashore", "sound effects"),
    (124, "Bird Tweet", "sound effects"),
    (125, "Telephone Ring", "sound effects"),
    (126, "Helicopter", "sound effects"),
    (127, "Applause", "sound effects"),
    (128, "Gunshot", "sound effects")
}

percussion = {
    (35, "Bass Drum 2", "drumset"),
    (36, "Bass Drum 1", "drumset"),
    (37, "Side Stick/Rimshot", "drumset"),
    (38, "Snare Drum 1", "drumset"),
    (39, "Hand Clap"),
    (40, "Snare Drum 2", "drumset"),
    (41, "Low Tom 2", "drumset"),
    (42, "Closed Hi-hat", "drumset"),
    (43, "Low Tom 1", "drumset"),
    (44, "Pedal Hi-hat", "drumset"),
    (45, "Mid Tom 2", "drumset"),
    (46, "Open Hi-hat", "drumset"),
    (47, "Mid Tom 1", "drumset"),
    (48, "High Tom 2", "drumset"),
    (49, "Crash Cymbal 1", "drumset"),
    (50, "High Tom 1", "drumset"),
    (51, "Ride Cymbal 1", "drumset"),
    (52, "Chinese Cymbal"),
    (53, "Ride Bell"),
    (54, "Tambourine"),
    (55, "Splash Cymbal", "drumset"),
    (56, "Cowbell"),
    (57, "Crash Cymbal 2", "drumset"),
    (58, "Vibra Slap"),
    (59, "Ride Cymbal 2", "drumset"),
    (60, "High Bongo"),
    (61, "Low Bongo"),
    (62, "Mute High Conga"),
    (63, "Open High Conga"),
    (64, "Low Conga"),
    (65, "High Timbale"),
    (66, "Low Timbale"),
    (67, "High Agogô", "High Agogo"),
    (68, "Low Agogô", "Low Agogo"),
    (69, "Cabasa"),
    (70, "Maracas"),
    (71, "Short Whistle"),
    (72, "Long Whistle"),
    (73, "Short Güiro", "Short Guiro"),
    (74, "Long Güiro", "Long Guiro"),
    (75, "Claves"),
    (76, "High Wood Block"),
    (77, "Low Wood Block"),
    (78, "Mute Cuíca", "Mute Cuica"),
    (79, "Open Cuíca", "Open Cuica"),
    (80, "Mute Triangle"),
    (81, "Open Triangle")
}
